/*
Copyright 2005, 2006, 2007 Dennis van Weeren

This file is part of Minimig

Minimig is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3 of the License, or
(at your option) any later version.

Minimig is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

This is a simple FAT16 handler. It works on a sector basis to allow fastest acces on disk
images.

Simplified from Minimig version

*/

#include <stdio.h>
#include <string.h>
#include <stdbool.h>

#include "mmc.h"
#include "fat16.h"
#include "fat16_defs.h"

/* Enable / Disable Debug info */
#undef DEBUG_FAT

/* Structure containing selected partition data */
struct partition_info {
	unsigned char	fatType;	// 0x10=FAT16, 0x20=FAT32
	unsigned long	partStart;	// start LBA of partition
	unsigned long	fatStart;	// start LBA of first FAT table
	unsigned long	dataStart;	// start LBA of data field
	unsigned long	rootDirCluster;	// root directory cluster 
	unsigned long	rootDirStart;	// start LBA of root directory table
	unsigned short	rootDirEntries;	// start LBA of root directory table
	unsigned char	fatNo;		// number of FAT tables
	unsigned char	clusterSize;	// size of a cluster in blocks
	unsigned long	clusterMask;	// binary mask of cluster number
};


/* Open file info */
struct file_info {
	unsigned char	name[12];  	// Short file name
	unsigned char	attributes;	// File attributes
	unsigned short	entry;		// File-entry index in directory table
	unsigned long	sector;  	// Sector index in file
	unsigned long	len;		// File size
	unsigned long	cluster;	// Current cluster
	unsigned long	firstCluster;	// First file cluster
};


#define MAX_LFN_SIZE	128				// Maximum supported long file name, default is 256

/* Directory Seek Modes */
#define DIRECTORY_BROWSE_START		0		// start search from beginning of directory
#define	DIRECTORY_BROWSE_NEXT		2		// find next file in directory

/* Sector buffer. Shared globally to save space on PIC */
static unsigned long cached_cur_lba;

/* Selected Partition */
static struct partition_info fat_partition;

static unsigned char longFilename[MAX_LFN_SIZE];	// Default long filename entry

static struct file_info fat_cur_file;			// global file handle
static struct file_info fat_cur_dir;			// global directory file handle

// Long filename entry char pos
static const unsigned char charLFNPos[] = { 1, 3, 5, 7, 9, 14, 16, 18, 20, 22, 24, 28, 30 };


#ifdef DEBUG_FAT
#define pr_debug(args...)	printf_tiny(args)
#else
#define pr_debug(args...)
#endif

/* Caching functions for MMC */
static unsigned char cached_read(unsigned long lba)
{
	if (cached_cur_lba == lba)
		return true;

	if (!mmc_read(lba, secbuf))
		return false;

	cached_cur_lba = lba;
	return true;
}

static unsigned char cached_write(unsigned long lba)
{
	cached_cur_lba = lba;
	return mmc_write(lba, secbuf);
}

/* Calculate LBA from current file position cluster and sector */
static unsigned long fat_get_lba(struct file_info near *file)
{
	unsigned long lba;

	/* Calc sector to read, when first cluster is 0 it is root directory of FAT16 */
	if (file->firstCluster == 0) {
		lba = fat_partition.rootDirStart;
		lba += file->sector;
	} else {
		lba = fat_partition.dataStart;				// start of data in partition
		lba += (file->cluster-2) * fat_partition.clusterSize;	// cluster offset
		lba += file->sector & ~fat_partition.clusterMask;	// sector offset in cluster
	}
	return lba;
}

/* Find next file cluster from the FAT */
static unsigned char fat_get_next_cluster(struct file_info near *file)
{
	unsigned short index;
	unsigned long fatSectorLba;

	/* calculate sector that contains FAT-link */
	fatSectorLba = fat_partition.fatStart;

	if (fat_partition.fatType & 0x10) {
		/* FAT16 256 cluster infos in sector on disk */
		fatSectorLba += (file->cluster) >> 8;

		/* calculate offset */
		index = (file->cluster) & 0xFF;
		index <<= 1;
	} else {
		/* FAT32 128 cluster infos in sector on disk */
		fatSectorLba += (file->cluster) >> 7;

		/* FAT32 128 cluster infos in sector */
		index = (file->cluster) & 0x7F;
		index <<= 2;
	}

	/* Read sector of FAT */
	if (!cached_read(fatSectorLba))
		return false;

	if (fat_partition.fatType & 0x10) {
		/* get FAT-link */
		file->cluster = *((unsigned short*)&secbuf[index]);

		/* Exit if end of chain */
		if (file->cluster == 0xffff)
			return false;
	} else {
		/* For FAT32 Read long instead short */
		file->cluster = *((unsigned long*)&secbuf[index]);
		file->cluster &= 0x0FFFFFFF;

		/* Exit if end of chain */
		if (file->cluster == 0x0fffffff)
			return false;
	}
	return true;
}

/* Point to next sector in file */
static unsigned char fat_do_next_sector(struct file_info near *file)
{
	/* increment sector index */
	file->sector++;

	/* if we are now in another cluster, look up cluster */
	if ((file->sector & ~fat_partition.clusterMask) == 0)
		if (!fat_get_next_cluster(file))
			return false;

	pr_debug(" n s=0x%lx c=%lx\r\n", file->sector, file->cluster);

	return true;
}

/* Seek within a file (or directory) entry */
static unsigned char fat_file_do_seek(struct file_info near *file, unsigned long sector)
{
	long currentClusterNo;
	long clusterNo;

	/* Calculate Current ClusterNo to avoid seek if new secor is on
	 * same cluster
	 */
	currentClusterNo = (long)file->sector;
	currentClusterNo /= (long)fat_partition.clusterSize;
	
	/* Sector in file to read */
	file->sector = sector;

	/* Calculate new ClusterNO in file */
	clusterNo = (long)file->sector;
	clusterNo /= (long)fat_partition.clusterSize;

	/* Calc ClusterNo Difference */
	currentClusterNo -= clusterNo; 
	
	/* Check if we are on same cluster */
	if (currentClusterNo == 0) {
		pr_debug(" S s=0x%lx c=%lx\r\n", file->sector, file->cluster);
		return true;
	}

	/* Reset current cluster in file to first */
	file->cluster = file->firstCluster;
		
	/* If first cluster in file exit */
	if (clusterNo == 0)
		return true;

	/* Loop through cluster links */
	while(--clusterNo >= 0) {
		if (!fat_get_next_cluster(file))
			return false;
		//		pr_debug(" -seek: clusterNo: 0x%lx", clusterNo);
		//		pr_debug(" , cluster: 0x%lx\r\n", file->cluster);
	}

	pr_debug(" s s=0x%lx c=%lx\r\n", file->sector, file->cluster);
	return true;
}

static unsigned char fat_process_dir_entry(struct file_info near *file, union FAT_directoryEntry * dirEntry)
{
	short i;
	short char_offset;

	/* Check if we are at directory end */
	if (dirEntry->entry.shortName.name[0] == FAT_ENTRY_FREE) {
		pr_debug("Entry free %d\r\n", file->entry);
		return 0;	
	}

	/* Check if file deleted */
	if (FAT_ENTRY_DELETED != dirEntry->entry.shortName.name[0]) {	
		/* Check if longfilename */
		if(FAT_ATTRIB_LFN_TEXT == dirEntry->entry.attributes) {
			/* Calc pos to copy part of LFN */
			char_offset = ((dirEntry->LFN.sequenceNo & 0x3f)-1) * 13;
	        
			/* Copy part of LFN */
			for(i = 0; i < 13 && (char_offset + i) < (MAX_LFN_SIZE - 1); i++)
				longFilename[char_offset + i] = dirEntry->bytes[charLFNPos[i]];
		} else if (dirEntry->entry.attributes &
			   (FAT_ATTRIB_HIDDEN | FAT_ATTRIB_SYSTEM | FAT_ATTRIB_VOLUME)) {
			pr_debug("FAT: Skiping entry no:%d attr:0x%02X\n",
				 file->entry, dirEntry->entry.attributes);
			/* Clear longfile name just in case */
			memset(longFilename,0,MAX_LFN_SIZE);
		} else {
			/* Copy name */
			memcpy(&file->name, &dirEntry->entry.shortName, 11);
			file->name[11] = 0x00; /* In theory not needed since never trashed */
			file->attributes = dirEntry->entry.attributes;
			file->len = dirEntry->entry.length; /* get length of file */
			file->firstCluster = (unsigned long)dirEntry->entry.firstClusterLow;	// get first cluster of file
			/* FAT32 high bytes of cluster */
			if (fat_partition.fatType & 0x20)
				file->firstCluster |= ((unsigned long)dirEntry->entry.firstClusterHigh) << 16;
			file->cluster = file->firstCluster; /* Copy current cluster to first cluster */
			file->sector = 0; /* Reset current sector index */

			/* Check if there is no LFN than copy short name as LFN */
			/* Skip entries that have LFN and entries starting with '.' (Parent and Current dir) */
			if(!longFilename[0] && (file->name[0] != '.')) {
				char_offset = 0;
				for(i=0; i < 11; i++) {
					if(file->name[i] != ' ')
						longFilename[char_offset++] = file->name[i];
					if(7 == i)
						longFilename[char_offset++] = '.';
				}
				/* Delete Last dot */
				if(longFilename[char_offset-1]=='.')
					char_offset--;
				longFilename[char_offset] = 0x00;
			}
			return 1;			
		}
	}
	return 2;
}

/* Iterate directory entries */
static unsigned char fat_get_dir_entry(struct file_info near *file, struct file_info near *dir,
				       unsigned char mode)
{
	short entryOffset;
	char rc;

	/* Clear long file name */
	memset(longFilename,0,MAX_LFN_SIZE);

	/* Check for iteration mode */
	switch(mode) {
	case DIRECTORY_BROWSE_NEXT:
		file->entry++;
		break;
	case DIRECTORY_BROWSE_START:
	default:
		file->entry = 0;
		break;
	}
	
	/* Seek directory to selected file entry */
	if(!fat_file_do_seek(dir, file->entry >> 4))
		return false;

	/* Infinite loop for search */
	while(1) {
		/* Read sector in buffer */
		if (!cached_read(fat_get_lba(dir)))
			return false;

		/* Calculate file entry offset in directory sector, 16 entries in sector 32 byte big */
		entryOffset = (file->entry & 0xF) << 5;

		/* Process directory entry */
		rc = fat_process_dir_entry(file, (union FAT_directoryEntry *)(secbuf + entryOffset));
		if (!rc)
			return false;
		if (rc == 1)
			return true;

		/* Go to next Fat Entry */
		file->entry++;
		
		/* Check should we seek directory to next sector */
		if (!(file->entry & 0xF))
			if (!fat_do_next_sector(dir))
				return false;
	}
}


/* Move to root directory */
static void fat_cd_root_dir(struct file_info near *dir)
{
	/* Clear Entry */
	memset(dir, 0, sizeof(struct file_info));
	dir->name[0] = '/';
	dir->attributes = FAT_ATTRIB_DIR;
	dir->firstCluster = fat_partition.rootDirCluster;
	dir->cluster = dir->firstCluster;
}

/* Find file by name and open it, return len or -1 */
long fat_open(const unsigned char *name)
{
	struct file_info near *file = &fat_cur_file;
	unsigned char mode = DIRECTORY_BROWSE_START;
	unsigned char i;

	pr_debug("FAT: Open file: %s\n",name);
	
	while (fat_get_dir_entry(file, &fat_cur_dir, mode)) {
		pr_debug(" found: %s...",file->name);
		for(i=0; i<11; i++)
			if (file->name[i]!=name[i])
				break;
		if (i==11) {
			pr_debug(" GOOD !\n"); 
			return file->len;
		}
		pr_debug("\n"); 
		mode = DIRECTORY_BROWSE_NEXT;
	}
	
	pr_debug("FAT: File not found\n");

	return -1;
}

unsigned char fat_file_read(void)
{
	return cached_read(fat_get_lba(&fat_cur_file));
}

unsigned char fat_file_write(void)
{
	return cached_write(fat_get_lba(&fat_cur_file));
}

unsigned char fat_file_seek(unsigned long sector)
{
	return fat_file_do_seek(&fat_cur_file, sector);
}
 
unsigned char fat_file_next_sector(void)
{
	return fat_do_next_sector(&fat_cur_file);
}
 
void fat_inval_cache(void)
{
	cached_cur_lba = 0xffffffff;
}

/* fat_init()
 *
 * Checks if a card is present. if a card is present it will check for
 * a valid FAT16 or FAT32 primary partition
 */
unsigned char fat_init(void)
{
	unsigned long fatsize;	// size of fat
	unsigned long dirsize;	// size of directory region in sectors

	struct MBR_Disk *mbr = (struct MBR_Disk *)secbuf;
	struct FAT_Boot_Sector *boot = (struct FAT_Boot_Sector *)secbuf;

	/* Invalidate cache */
	cached_cur_lba = 0xffffffff;

	/* Read partition sector */
	if (!cached_read(0)) {
		pr_debug("FAT: Error Unable to read partition sector\n");
		return false;
	}
	
	/* Check for signature */
	if (mbr->signature != 0xAA55) {	
		pr_debug("FAT: Error invalid FAT signature\n");
		return false;
	}

	/* Check first partition type */
	switch (mbr->partitions[0].Type) {
		case PARTITION_TYPE_FAT16_32MB:
		case PARTITION_TYPE_FAT16:
			/* First partition filesystem type: FAT16 */
			fat_partition.fatType = 0x10;
			/* Get start of first partition */
			fat_partition.partStart = mbr->partitions[0].LBAFirst;
			break;

		case PARTITION_TYPE_FAT32:
		case PARTITION_TYPE_FAT32_LBA:
			/* First partition filesystem type: FAT32 */
			fat_partition.fatType = 0x20;
			/* get start of first partition */
			fat_partition.partStart = mbr->partitions[0].LBAFirst;
			break;
			
		default:
			pr_debug("FAT: Error no supported partition found\n");
			return false;
	}

	/* Read boot sector */
	if (!cached_read(fat_partition.partStart)) {	
		pr_debug("FAT: Error unable to read boot sector\n");
		return false;
	}

	/* Check for near-jump or short-jump opcode */
	if (0xE9 != boot->jumpInstruction[0] && 0xEB != boot->jumpInstruction[0]) {	
		pr_debug("FAT: Error invalid 0x%x boot sector jump\n",
			 boot->jumpInstruction[0]);
		return false;
	}

	/* Check if blocksize is really 512 bytes */
	if (0x200 != boot->bytesPerSector) {
		pr_debug("FAT: Error block size not 512 bytes it is 0x%x\n",
			 boot->bytesPerSector);
		return false;
	}

	/* Check medium descriptor byte, must be 0xf8 for hard drive */
	if (0xF8 != boot->mediaDescriptor) {
		pr_debug("Error invalid media descriptor\n");
		return false;
	}
	
	/* Get cluster size */
	fat_partition.clusterSize = boot->sectorsPerCluster;
	/* calculate cluster mask */
	fat_partition.clusterMask = ~(fat_partition.clusterSize-1);

	
	/* Calculate drive's parameters from bootsector, first up is size of directory */
	/* Get Max ROOT Dir entries FAT16 only !!*/
	fat_partition.rootDirEntries = boot->maxRootEntries;

	/* Calculate start of FAT, size of FAT and number of FAT's */
	fat_partition.fatStart = fat_partition.partStart + boot->reservedSectorCount;
	fat_partition.fatNo = boot->noOfFATs;

	/* When FAT16 */ 
	if (0x10 == fat_partition.fatType) {
		fatsize = boot->sectorsPerFAT;
		// Fat 16 Root dir cluster
		fat_partition.rootDirCluster = 0;
		// calculate start of FAT16 ROOT directory 
		fat_partition.rootDirStart = fat_partition.fatStart + (fat_partition.fatNo * fatsize);
		// Calculate dire size in sectors
		dirsize = ((fat_partition.rootDirEntries<<5)+511)>>9;
		// calculate start of data
		fat_partition.dataStart = fat_partition.rootDirStart + dirsize;
	} else {
		fatsize = boot->extParams.fat32Ext.sectorsPerFAT;
		// calculate data start
		fat_partition.dataStart = fat_partition.fatStart + (fat_partition.fatNo * fatsize);
		// Fat 32 Root dir cluster
		fat_partition.rootDirCluster = boot->extParams.fat32Ext.rootDirCluster; 
		// Get FAT32 root dir start 
		fat_partition.rootDirStart = fat_partition.dataStart +
			(fat_partition.rootDirCluster-2) * fat_partition.clusterSize;
	}

	/* Open Root Directory */
	fat_cd_root_dir(&fat_cur_dir);
	
	pr_debug("FAT%d partition found.\n", fat_partition.fatType);
	pr_debug("Fat Size:	%ld\n",fatsize);
	pr_debug("No of fats:	%d\n",fat_partition.fatNo);
	pr_debug("Fat start:	%ld\n", fat_partition.fatStart);
	pr_debug("Direntries:	%d\n",fat_partition.rootDirEntries);
	pr_debug("Dir start:	%ld\n", fat_partition.rootDirStart);
	pr_debug("Data start:	%ld\n", fat_partition.dataStart);
	pr_debug("Cluster size:	%d\n",fat_partition.clusterSize);
	pr_debug("Cluster mask:	%lx\n",fat_partition.clusterMask);

	return true;
}

