/*
Copyright 2005, 2006, 2007 Dennis van Weeren

This file is part of Minimig

Minimig is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3 of the License, or
(at your option) any later version.

Minimig is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

Hardware control routines

27-11-2005	-started coding
29-01-2006	-done a lot of work
31-01-2006	-added key repeat
06-02-2006	-took out all button handling stuff

-- Goran Ljubojevic --
2009-11-13	- OsdCommand added
2009-11-21	- small cleanup
2009-12-20	- systimer reset on every hardware init to support propper timings on reset
2009-12-30	- Support for new FPGA firmware added in header file
			- GetFPGAStatus function added
2010-01-29	- ResetFPGA() macro added to header file.
2010-08-21	- YQ100818 FPGA core support
2010-08-26	- Added firmwareConfiguration.h
2010-09-09	- Added _SPI macro to help save space on multiple SPI calls
2010-10-15      - Moved over to minimigmac, coding style changes, etc...
*/

#include <stdio.h>

#include "hardware.h"

static volatile unsigned short systimer;

/*initialize hardware*/
void hardware_init(void)
{
	/* Disable analog inputs */
	ADCON1 = 0b00000110;

	/* Initalize output register */
	PORTA = 0b00100011;
	PORTB = 0b00000000;
	PORTC = 0b00010001;

	/* Enable PORTB weak pullup */
	INTCON2bits.RBPU = 0;

	/* Initialize SPI to master mode 0, sample on posedge, 5Mhz */
	SSPSTAT = 0x40;
	SSPCON1 = 0x20;

	/* Initialize input/ouput configuration */
	TRISA = 0b11001100;
	TRISB = 0b00101011;
	TRISC = 0b10010000;

	/* Initialize serial port */
	SPBRG = 10;     /*115200 BAUD @ 20MHz*/
	TXSTA = 0x24;
	RCSTA = 0x90;

	/* Init timer0, internal clk, prescaler 1:256 */
	T0CON = 0xc7;
	TMR0H = 0;
	TMR0L = 0;

	/* Enable interrupt for timer 0 */
	INTCONbits.TMR0IE = 1;
	INTCONbits.GIE = 1;
}

/* interrupt service routine */
static void isr(void) interrupt 1
{
	unsigned short t;
	static unsigned char button_state;

	/* Clear timer 0 interrupt flag */
	INTCONbits.TMR0IF = 0;

	/* Set timer to timeout every 10ms
	 *  @20Mhz --> instruction = 200ns
	 *  200ns * 256 * 195  =  10ms
	 *
	 * Not sure we can just cast, so for now be gross
	 * since SDCC doesn't give us a 16-bit TMR0
	 */
	t = TMR0L;
	t |= ((unsigned short)TMR0H) << 8;
	t -= 195;
	TMR0H = t >> 8;
	TMR0L = t & 0xff;

	/* Increment system timer */
	systimer++;

	/* Debounce button */
	if (check_button()) {
		if (button_state < 10) {
			button_state++;
			if (button_state == 10) {
				putchar('[');
				putchar('N');
				putchar('M');
				putchar('I');
				putchar(']');
				NMI = 1;
			}
		}
	} else {
		button_state = 0;
		NMI = 0;
	}
}

/* Get system timer + offset (handy for lots of things)*/
unsigned short get_timer(unsigned short offset)
{
	unsigned short r;

	/* Get system time SAFELY */
	INTCONbits.GIE = 0;
	r = systimer;
	INTCONbits.GIE = 1;

	/*add offset*/
	r += offset;

	return r;
}

/* Check if timer is past given time in <t>
 * it may be maximum 30000 ticks in the future
 */
char check_timer(unsigned short t)
{
	/* calculate difference */
	INTCONbits.GIE = 0;
	t -= systimer;
	INTCONbits.GIE = 1;

	/* check if <t> has passed */
	return t > 30000;
}

void wait_timer(unsigned short time)
{
    time = get_timer(time);
    while (!check_timer(time))
	    ;
}

/* SPI-bus. Should inline ? */
unsigned char do_spi(unsigned char d)
{
	SSPBUF = d;
	while (!SSPSTATbits.BF);	/*Wait untill controller is ready*/
	return(SSPBUF);			/*Return with received value*/	
}

/* FPAG configuration serial interface */
void shift_fpga(unsigned char byte)
{
	/*bit 0*/
	DIN = 0;
	CCLK = 0;
	if (byte & 0x80)
		DIN = 1;
	CCLK = 1;

	/*bit 1*/
	DIN = 0;
	CCLK = 0;
	if (byte & 0x40)
		DIN = 1;
	CCLK = 1;

	/*bit 2*/
	DIN = 0;
	CCLK = 0;
	if (byte & 0x20)
		DIN = 1;
	CCLK = 1;

	/*bit 3*/
	DIN = 0;
	CCLK = 0;
	if (byte & 0x10)
		DIN = 1;
	CCLK = 1;

	/*bit 4*/
	DIN = 0;
	CCLK = 0;
	if (byte & 0x08)
		DIN = 1;
	CCLK = 1;

	/*bit 5*/
	DIN = 0;
	CCLK = 0;
	if (byte & 0x04)
		DIN = 1;
	CCLK = 1;

	/*bit 6*/
	DIN = 0;
	CCLK = 0;
	if (byte & 0x02)
		DIN = 1;
	CCLK = 1;

	/*bit 7*/
	DIN = 0;
	CCLK = 0;
	if (byte & 0x01)
		DIN = 1;
	CCLK = 1;
}

/* Put out a chacter to the serial port */
void putchar(char ch) __wparam
{
	while(TXSTAbits.TRMT == 0);
	TXREG = ch;
	if (ch == 10) {
		while(TXSTAbits.TRMT == 0);
		TXREG = 13;
	}
}






