/*
 * MMC/SD-Card driver
 *
 * Essentially copied from Minimig
 *
 * Current implementation is polled & synchronous
 */

#include <stdio.h>

#include "hardware.h"
#include "mmc.h"
#include "mmc_cmds.h"

/* Enable / Disable Debug info */
#undef DEBUG_SDMMC

/* MMC, or SD Card Specifification V1, V2, SDHC */
#define	MMC_CARD_NONE			0x00
#define	MMC_CARD_MMC			0x01
#define	MMC_CARD_SD_RAW_SPEC_1		0x02
#define MMC_CARD_SD_RAW_SPEC_2		0x04
#define MMC_CARD_SD_RAW_SPEC_SDHC	0x08

static unsigned char mmc_card;
static unsigned char mmc_crc_7;
static unsigned int mmc_timeout;
static unsigned char mmc_resp[5];

#ifdef DEBUG_SDMMC
#define pr_debug(args...)	printf_tiny(args)
#else
#define pr_debug(args...)
#endif

/* Calculate CRC7 checksum */
static void mmc_add_crc7(unsigned char c)
{
	unsigned char i;

	for (i = 0; i < 8; i++) {
		mmc_crc_7 <<= 1;
		if (c & 0x80)
			mmc_crc_7 ^= 0x09;
		if (mmc_crc_7 & 0x80)
			mmc_crc_7 ^= 0x09;
		c <<= 1;
	}
}

/* Send a byte and calculate CRC */
static void mmc_spi_crc(unsigned char cmd)
{
	do_spi(cmd);
	mmc_add_crc7(cmd);
}

/* Send a command to the SDcard */
static void mmc_cmd_r0(char cmd, unsigned short addr_h, unsigned short addr_l)
{
	mmc_crc_7 = 0;

	/* Flush SPI-bus */
	do_spi(0xff);

	/* Send command */
	mmc_spi_crc(cmd);

	/* Send address */
	mmc_spi_crc(addr_h >> 8);
	mmc_spi_crc(addr_h & 0xff);
	mmc_spi_crc(addr_l >> 8);
	mmc_spi_crc(addr_l & 0xff);

	/* Cook and send CRC */
	mmc_crc_7 <<= 1;
	mmc_crc_7++;
	do_spi(mmc_crc_7);
}

/* Send a command to the SDcard, a one byte response is expected */
static void mmc_cmd_r1(char cmd, unsigned short addr_h, unsigned short addr_l)
{
	unsigned char i = 100;

	mmc_cmd_r0(cmd, addr_h, addr_l);
	do
		mmc_resp[0] = do_spi(0xff);
	while (mmc_resp[0] == 0xff && --i);
}

/* Send a command to the SDcard, a two byte response is expected */
static void mmc_cmd_r2(char cmd, unsigned short addr_h, unsigned short addr_l)
{
	unsigned char i = 100;

	mmc_cmd_r0(cmd, addr_h, addr_l);
	do
		mmc_resp[0] = do_spi(0xff);
	while (mmc_resp[0] == 0xff && --i);
	mmc_resp[1] = do_spi(0xff);
}

/* Send a command to the SDcard, a five byte response is expected */
static void mmc_cmd_r3(char cmd, unsigned short addr_h, unsigned short addr_l)
{
	unsigned char i = 100;

	mmc_cmd_r0(cmd, addr_h, addr_l);
	do
		mmc_resp[0] = do_spi(0xff);
	while (mmc_resp[0] == 0xff && --i);
	mmc_resp[1] = do_spi(0xff);
	mmc_resp[2] = do_spi(0xff);
	mmc_resp[3] = do_spi(0xff);
	mmc_resp[4] = do_spi(0xff);
}

/* Enable the MMC/SD card correctly */
unsigned char mmc_init(void)
{
	unsigned short lp;

	/* Set SPI clock .. Might need to be smarter at changing the
	 * SPI clock based on what device is enabled...
	 */
	SSPCON1 = 0x22; //spiclk = Fosc/64 (init clock 100-400 kHz)

	_M_CD = 0;
	_M_CS = 1;
	wait_timer(10);

	_M_CD = 1; /* Enable clock*/
	_M_CS = 1; /* SDcard Disabled */
	
	/* Set SDcard in SPI-Mode, Reset*/
	/* (10 * 8bits = 80 clockpulses) */
	for (lp = 0; lp < 10; lp++)
		do_spi(0xff);

	/* Delay for a lot of milliseconds (least 16 bus clock cycles) */
	for (lp=0; lp < 56000; lp++)
		; /* PRAY sdcc doesn't optimize that out !
		     Should use timer instead ! */
	wait_timer(10);

	/* SDcard Enabled */
	_M_CS = 0;

	/* No card detected by default */
	mmc_card = MMC_CARD_NONE;
	
	/* CMD0: Reset all cards to IDLE state */
	mmc_cmd_r1(CMD0, 0, 0);
	if (mmc_resp[0] != 0x01) {
		pr_debug("No card detected ! (resp=%x)\n", mmc_resp[0]);
		goto fail;
	}

	/* Default to MMC */
	mmc_card = MMC_CARD_MMC;

	/* Check for SD Card V2 */
	/* (Voltage 2.7V - 3.6V, 0xaa test pattern) */
	mmc_cmd_r3(CMD8, 0x0000, 0x01aa);
	if (!(mmc_resp[0] & 0x04)) {
		if (mmc_resp[3] == 0x01 && mmc_resp[4] == 0xaa) {
			pr_debug("SD card V2 detected, possible SDHC\n");
			mmc_card |= MMC_CARD_SD_RAW_SPEC_2;
		} else {
			pr_debug("Error detecting SD V2 card\n");
			goto fail;
		}
	} else {
		/* Check for SD card V1 */
		mmc_cmd_r1(CMD55, 0, 0);
		mmc_cmd_r1(CMD41, 0, 0);
		if (!(mmc_resp[0] & 0x04)) {
			pr_debug("SD card V1 detected\n");
			mmc_card |= MMC_CARD_SD_RAW_SPEC_1;
		} else
			pr_debug("MMC-card detected\n");
	}
	
	/* Wait for card to get ready */
	mmc_timeout = 0;
	while (1) {
		if (mmc_card & (MMC_CARD_SD_RAW_SPEC_1 | MMC_CARD_SD_RAW_SPEC_2)) {
			/* When SD Card detected */
			lp = 0x0;
			if (mmc_card & MMC_CARD_SD_RAW_SPEC_2)
				lp = 0x4000;
			/* Activate SD card init process */
			mmc_cmd_r1(CMD55, 0, 0);
			mmc_cmd_r1(CMD41, lp, 0);
		} else {
			/* Activate the MMC cards init process */
			mmc_cmd_r1(CMD1, 0, 0);
		}
		
		/* Check if card is idle */
		if (mmc_resp[0] == 0x00)
			break;
		mmc_timeout++;
		if (mmc_timeout == 1000) {
			pr_debug("SD/MMC ACMD41/CMD1 response timeout...\n");
			goto fail;
		}
	}

	/* Check for SDHC card */
	if (mmc_card & MMC_CARD_SD_RAW_SPEC_2) {
		/* Get operating conditions */
		mmc_cmd_r3(CMD58, 0, 0);
		if (mmc_resp[0]) {
			pr_debug("SD/MMC: Failed to get operating conditions\n");
			goto fail;
		}
		
		if (mmc_resp[1] & 0x40) {	
			pr_debug("SDHC Card Detected\n");
			mmc_card |= MMC_CARD_SD_RAW_SPEC_SDHC;
		}
	}

	/* Set block size to 512 bytes */
	mmc_timeout = 0;
	while(1) {
		mmc_cmd_r1(CMD16, 0x0000, 0x0200);
		if (!mmc_resp[0])
			break;
		mmc_timeout++;
		if(mmc_timeout == 100) {
			pr_debug("Set block size to 512 timeout\n");
			goto fail;
		}
	}
	
	spi_disable_sdcard();
	
	/* XXX Stuff below from minimig... it's strange... it sets SSPCON1 to a
	 * different speed from what is done in hardware_init.c for everybody
	 */
	// TODO: Change speed for card access
	SSPCON1 = 0x20; //spiclk = Fosc/4 (5 MHz)
	return true;
 fail:
	mmc_card = MMC_CARD_NONE;
	spi_disable_sdcard();
	return false;
}

/* Read single block (with block-size set by CMD16 to 512 by default) */
unsigned char mmc_read(unsigned long lba, unsigned char *out_data)
{
	unsigned short upper_lba, lower_lba;
	unsigned char i;
	unsigned char *p;

	/* !SDHC uses byte address instead of LBA */
	if (!(mmc_card & MMC_CARD_SD_RAW_SPEC_SDHC))
		lba <<= 9;

	pr_debug("SD/MMC: Reading LBA 0x%lx\n", lba);
	
	upper_lba = (unsigned short)(lba>>16);
	lower_lba = (unsigned short)lba;

	spi_enable_sdcard();

	mmc_cmd_r1(CMD17, upper_lba, lower_lba);

	/* Exit if invalid response*/
	if (mmc_resp[0]) {
		pr_debug("SD/MMC CMD17: invalid response %x\n", mmc_resp[0]);
		goto fail;
	}

	/* Wait for start of data transfer with timeout */
	mmc_timeout = 0;
	while(do_spi(0xff) != 0xfe) {
		if (mmc_timeout++ >= 50000) {
			pr_debug("SD/MMC CMD17: no data token\n");
			goto fail;
		}
	}

	/* Read data and exit OK */
	p = out_data;
	for (i = 0; i < 128; i++) {
		SSPBUF = 0xff;
		while (!SSPSTATbits.BF)
			;
		*(p++) = SSPBUF;
		SSPBUF = 0xff;
		while (!SSPSTATbits.BF)
			;
		*(p++) = SSPBUF;
		SSPBUF = 0xff;
		while (!SSPSTATbits.BF)
			;
		*(p++) = SSPBUF;
		SSPBUF = 0xff;
		while (!SSPSTATbits.BF)
			;
		*(p++) = SSPBUF;
	}

	/* XXX Todo: check CRC */
	do_spi(0xff);	//Read CRC lo byte
	do_spi(0xff);	//Read CRC hi byte

	spi_disable_sdcard();
	return true;
 fail:
	spi_disable_sdcard();
	return false;
}



/* Write: 512 Byte-Mode, this will not work (read MMC and SD-card specs) with any other
 * sector/block size
 */
unsigned char mmc_write(unsigned long lba, unsigned char *in_data)
{
	unsigned short upper_lba, lower_lba;
	unsigned char i;
	unsigned char *p;

	/* !SDHC uses byte address instead of LBA */
	if (!(mmc_card & MMC_CARD_SD_RAW_SPEC_SDHC))
		lba <<= 9;

	pr_debug("SD/MMC: Writing LBA 0x%lx\n", lba);

	upper_lba = (unsigned short)(lba>>16);
	lower_lba = (unsigned short)lba;

	spi_enable_sdcard();

	mmc_cmd_r1(CMD24, upper_lba, lower_lba);

	/* Exit if invalid response*/
	if (mmc_resp[0]) {
		pr_debug("SD/MMC CMD24: invalid response %x\n", mmc_resp[0]);
		goto fail;
	}

	do_spi(0xff);	//One byte gap
	do_spi(0xfe);	//Send Data token

	/* Send bytes for sector */
	p = in_data;
	for (i = 0; i < 128; i++) {
		SSPBUF = *(p++);
		while (!SSPSTATbits.BF)
			;
		SSPBUF = *(p++);
		while (!SSPSTATbits.BF)
			;
		SSPBUF = *(p++);
		while (!SSPSTATbits.BF)
			;
		SSPBUF = *(p++);
		while (!SSPSTATbits.BF)
			;
	}

	/* XXX Todo: Send proper CRC ? */
	do_spi(0xff);	//Send CRC lo byte
	do_spi(0xff);	//Send CRC hi byte

	/* Read packet response */
	i = do_spi(0xff);

	/* Status codes
	 *  010 = Data accepted
	 *  101 = Data rejected due to CRC error
	 *  110 = Data rejected due to write error
	 */
	i &= 0b00011111;
	if (i != 0b00000101) {
		pr_debug("SD/MMC CMD24: write error %x\n", i);
		goto fail;
	}

	mmc_timeout = 0;
	/* Wait until the card has finished writing the data */
	while (do_spi(0xff) == 0x00) {
		if (mmc_timeout++ >= 50000) {
			pr_debug("SD/MMC CMD24: busy wait timeout\n");
			goto fail;
		}
	}
	spi_disable_sdcard();
	return true;
 fail:
	spi_disable_sdcard();
	return false;
}

